import numpy as np
from py3gpp.nrPolarDecode import nrPolarDecode
from py3gpp.nrRateRecoverPolar import nrRateRecoverPolar
from py3gpp.nrCRCDecode import nrCRCDecode
from py3gpp.nrPBCHPRBS import nrPBCHPRBS


def nrBCHDecode(softbits, L, lssb=None, ncellid=None):
    K = 32
    E = 864
    L = 8
    N = 512
    matched = nrRateRecoverPolar(softbits, K, N, False, False)

    decoded = nrPolarDecode(matched, K, E, L)
    scrblk, crc_result = nrCRCDecode(decoded, "24C")

    if (lssb is None) or (ncellid is None):
        return scrblk, crc_result

    # descrambling according to TS38.212 7.1.2
    # fmt: off
    G = [16, 23, 18, 17, 8, 30, 10, 6, 24, 7, 0, 5, 3, 2, 1, 4, 9, 11, 12, 13, 14, 15, 19, 20, 21, 22, 25, 26, 27, 28,
         29, 31]
    # fmt: on
    SFN_PAYLOAD_BEGIN = 1
    SFN_PAYLOAD_LENGTH = 6
    SFN_2ND_LSB = SFN_PAYLOAD_LENGTH + 2
    SFN_3RD_LSB = SFN_PAYLOAD_LENGTH + 1
    v = 2 * scrblk[G[SFN_3RD_LSB]] + scrblk[G[SFN_2ND_LSB]]

    L_max = 8  # TODO: fix this
    A = 32
    if L_max in (4, 8):
        M = A - 3
    else:
        M = A - 6
    n = v * M
    tmp_seq = nrPBCHPRBS(ncellid, 0, len(scrblk) * 100)
    scrambling_seq = tmp_seq[n:][:A]
    scrambling_seq_final = np.zeros(A, "int")
    j = 0
    for i in range(A):
        is_ssb_idx = (i in (G[11], G[12], G[13])) and L_max == 64
        if is_ssb_idx or i == G[10] or i == G[SFN_2ND_LSB] or i == G[SFN_3RD_LSB]:
            scrambling_seq_final[i] = 0
        else:
            scrambling_seq_final[i] = scrambling_seq[j]
            j += 1
    a = np.bitwise_xor(scrambling_seq_final, scrblk)

    # deinterleaving according to TS38.212 7.1.1
    j_sfn = 0
    j_other = 14
    payload = np.empty(24, "int")
    for i in range(24):
        if (i >= SFN_PAYLOAD_BEGIN) and (i < (SFN_PAYLOAD_BEGIN + SFN_PAYLOAD_LENGTH)):
            payload[i] = a[G[j_sfn]]
            j_sfn += 1
        else:
            payload[i] = a[G[j_other]]
            j_other += 1
    lsbotfsfn = np.array([a[G[j_sfn]], a[G[j_sfn + 1]], a[G[j_sfn + 2]], a[G[j_sfn + 3]]])
    hrf = a[G[10]]
    msbidxoffset = 0  # TODO calculate this
    return scrblk, crc_result, payload, lsbotfsfn, hrf, msbidxoffset



from .PDSCHPTRSConfigBase import PDSCHPTRSConfigBase

class nrPDSCHPTRSConfig(PDSCHPTRSConfigBase):
    def __init__(self):
        super().__init__()
        self._PTRSPortSet = []

    @property
    def PTRSPortSet(self):
        raise NotImplementedError('not ready yet')
        return self._PTRSPortSet

    @PTRSPortSet.setter
    def PTRSPortSet(self, size):
        raise NotImplementedError('not ready yet')
        self._PTRSPortSet = size




# 38.211 7.4.3 SS/PBCH block

import numpy as np
from py3gpp.nrSymbolModulate import nrSymbolModulate
from py3gpp.nrPBCHPRBS import nrPBCHPRBS

def nrPBCH(ncellid, v, databits):
    n = len(databits)

    assert ncellid <= 1007 and ncellid >= 0
    assert v >= 0 and v <= 7
    assert n > 0

    # Scrambling, Section 7.3.3.1, TS 38.211
    prbs_bits = nrPBCHPRBS(ncellid, v, n)
    scrambled_bits = np.bitwise_xor(databits, prbs_bits)

    # Modulation, Section 7.3.3.2, TS 38.211
    res_sym = nrSymbolModulate(scrambled_bits, 'QPSK')

    return np.array(res_sym)




from .helper import _calc_m_seq
import numpy as np

# TS 38.211 Section 5.2.1
def nrPRBS(c_init, length):
    assert length > 0

    N = 31

    c_0 = np.zeros(N)
    c_0[0] = 1
    c_1 = np.zeros(N)
    for i in range(N):
        c_1[i] = (c_init>>i) & 1

    Nc = 1600
    seq_len = length + Nc
    mseq_0 = _calc_m_seq(N, c_0, [0, 3], seq_len)
    mseq_1 = _calc_m_seq(N, c_1, [0, 1, 2, 3], seq_len)

    gold_seq = (mseq_0 + mseq_1) % 2
    gold_seq = np.roll(gold_seq, -Nc)
    return gold_seq[: length]




class DMRSConfigBase():
     def __init__(self):
        self._NLayers = 1
        self._Ports = None
        self._DMRSTypeAPosition = 2
        self._DMRSAdditionalPosition = 0
        self._DMRSLength = 0
        self._CustomSymbolSet = []
        self._DMRSPortSet = []
        self._NIDNSCID = []
        self._NSCID = 0

     @property
     def NLayers(self):
        return self._NLayers

     @NLayers.setter
     def NLayers(self, N):
        assert N > 0
        self._NLayers = N

     @property
     def Ports(self):
        return self._Ports

     @property
     def DMRSTypeAPosition(self):
        return self._DMRSTypeAPosition

     @DMRSTypeAPosition.setter
     def DMRSTypeAPosition(self, pos):
        assert pos in [2, 3]
        self._DMRSTypeAPosition = pos

     @property
     def DMRSAdditionalPosition(self):
        return self._DMRSAdditionalPosition

     @DMRSAdditionalPosition.setter
     def DMRSAdditionalPosition(self, pos):
        assert pos in list(range(4))
        self._DMRSAdditionalPosition = pos

     @property
     def DMRSAdditionalPosition(self):
        return self._DMRSLength

     @DMRSAdditionalPosition.setter
     def DMRSAdditionalPosition(self, pos):
        assert pos in list(range(4))
        self._DMRSLength = pos

     @property
     def CustomSymbolSet(self):
        raise NotImplementedError('not ready yet')
        return self._CustomSymbolSet

     @CustomSymbolSet.setter
     def CustomSymbolSet(self, set):
        raise NotImplementedError('not ready yet')
        self._CustomSymbolSet = set

     @property
     def DMRSPortSet(self):
        raise NotImplementedError('not ready yet')
        return self._DMRSPortSet

     @DMRSPortSet.setter
     def DMRSPortSet(self, set):
        raise NotImplementedError('not ready yet')
        self._DMRSPortSet = set

     @property
     def NIDNSCID(self):
        return self._NIDNSCID

     @NIDNSCID.setter
     def NIDNSCID(self, nid):
        assert nid >= 0 and nid <= 65535, "The value must be 0..65535"
        self._NIDNSCID = nid

     @property
     def NSCID(self):
        return self._NSCID

     @NSCID.setter
     def NSCID(self, nscid):
        assert nscid in [0, 1], "The value must be 0 or 1"
        self._NSCID = nscid




import sys
import itertools
import numpy as np
import pytest

from py3gpp.nrPDSCH import nrPDSCH

sys.path.append("test_data")

from test_data.pdsch import pdsch_symbols_ref, pdsch_bits_ref

def run_nr_pdsch(cws, modulation, nlayers, nid, nrnti):
    pdsch_syms = nrPDSCH(cws, modulation, nlayers, nid, nrnti)
    pdsch_syms = np.around(pdsch_syms, 4)

    assert np.array_equal(pdsch_syms, pdsch_symbols_ref)

def test_run_nr_pdsch():
    run_nr_pdsch([pdsch_bits_ref], ["64QAM"], 1, 0, 0)



from .DMRSConfigBase import DMRSConfigBase

class nrPDSCHDMRSConfig(DMRSConfigBase):
    def __init__(self):
        super().__init__()
        self._DMRSConfigurationType = 1
        self._DMRSReferencePoint = 'CRB0'
        self._NumCDMGroupsWithoutData = 2
        self._DMRSDownlinkR16 = 0

    @property
    def DMRSConfigurationType(self):
        return self._DMRSConfigurationType

    @DMRSConfigurationType.setter
    def DMRSConfigurationType(self, size):
        assert size in [1, 2], "The value must be 1 or 2."
        self._DMRSConfigurationType = size

    @property
    def DMRSReferencePoint(self):
        raise NotImplementedError('not ready yet')
        return self._DMRSReferencePoint

    @DMRSReferencePoint.setter
    def DMRSReferencePoint(self, size):
        raise NotImplementedError('not ready yet')
        assert size in ['CRB0', 'PRB0'], "The value must be 'CRB0' or 'PRB0'."
        self._DMRSReferencePoint = size

    @property
    def NumCDMGroupsWithoutData(self):
        raise NotImplementedError('not ready yet')
        return self._NumCDMGroupsWithoutData

    @NumCDMGroupsWithoutData.setter
    def NumCDMGroupsWithoutData(self, N):
        raise NotImplementedError('not ready yet')
        assert N in [1, 2, 3], "The value must be 1, 2 or 3."
        self._NumCDMGroupsWithoutData = N

    @property
    def DMRSDownlinkR16(self):
        raise NotImplementedError('not ready yet')
        return self._DMRSDownlinkR16

    @DMRSDownlinkR16.setter
    def DMRSDownlinkR16(self, N):
        raise NotImplementedError('not ready yet')
        self._DMRSDownlinkR16 = N




# 38.211

import numpy as np

from py3gpp.nrPDSCHDMRS import PDSCHDMRSSyms
from py3gpp.configs.nrPDSCHConfig import nrPDSCHConfig

def nrPDSCHDMRSIndices(cfg: nrPDSCHConfig):
    frame_begin = cfg.NRBSize * min(cfg.PRBSet)
    frame_end = cfg.NRBSize * (max(cfg.PRBSet)+1)
    frame_size = cfg.NRBSize * cfg.NSizeBWP

    # Single frame DMRS positions
    dmrs_full_range = np.array(list(range(frame_begin, frame_end)))

    # Calculate DMRS positions in every occupied symbol. Frequency position
    if cfg.DMRS.DMRSConfigurationType == 1:
        # Takes every 2nd symbol
        occupied_res = dmrs_full_range[::2]
    elif cfg.DMRS.DMRSConfigurationType == 2:
        # Takes every 4th couple of symbols
        occupied_res = dmrs_full_range.reshape(-1, 2)[::3].ravel()

    # Calculates occupied symbols numbers. Time positions
    occupied_syms = PDSCHDMRSSyms(cfg)

    dmrs_indices = np.array([])
    for idx, sym in enumerate(occupied_syms):
        sym_offset = sym*frame_size
        dmrs_indices = np.append(dmrs_indices, (occupied_res + sym_offset))

    return dmrs_indices



from .PDSCHConfigBase import PDSCHConfigBase
from .CommonConfig import CommonConfig

class nrPDSCHConfig(PDSCHConfigBase, CommonConfig):
    def __init__(self):
        super().__init__()
        self._Nid = 0
        self._NSizeBWP = []
        self._NStartBWP = []

    @property
    def NRBSize(self):
        return 12

    @property
    def Nid(self):
        return self._Nid

    @Nid.setter
    def Nid(self, nid):
        assert nid >= 0 and nid <= 1007, "The value must be in the range 0..1007."
        self._Nid = nid

    @property
    def NSizeBWP(self):
        return self._NSizeBWP

    @NSizeBWP.setter
    def NSizeBWP(self, size):
        assert size > 1 and size < 275, "The value must be in the range 1...275."
        self._NSizeBWP = size

    @property
    def NStartBWP(self):
        return self._NStartBWP

    @NStartBWP.setter
    def NStartBWP(self, start):
        assert start >= 0 and start < 2473, "The value must be in the range 0...2473."
        self._NStartBWP = start



import numpy as np

def _calc_m_seq(N, c, taps, seq_len = None):
    if seq_len is None:
        seq_len = 2**N - 1
    m = np.zeros(seq_len, "int32")  # m-sequence
    assert len(c) == N, "c has wrong length"
    m[:N] = c
    for i in np.arange(0, seq_len - N):
        m[i + N] = np.sum(np.take(m, i + taps)) % 2
    return m


# TS 38.211 Section 5.2.1
def _calc_gold(c_init, length):
    N = 31

    c_0 = np.zeros(N)
    c_0[0] = 1
    c_1 = np.zeros(N)
    for i in range(N):
        c_1[i] = (c_init>>i) & 1

    Nc = 1600
    seq_len = length + Nc
    mseq_0 = _calc_m_seq(N, c_0, [0, 3], seq_len)
    mseq_1 = _calc_m_seq(N, c_1, [0, 1, 2, 3], seq_len)

    gold_seq = (mseq_0 + mseq_1) % 2
    gold_seq = np.roll(gold_seq, -Nc)
    return gold_seq[: length]




class CommonConfig():
    def __init__(self):
        self._MappingType = "A"
        self._SymbolAllocation = [0, 14]
        self._PRBSet = list(range(0, 52))
        self._RNTI = 1
        self._EnablePTRS = 0

    @property
    def MappingType(self):
        return self._MappingType

    @MappingType.setter
    def MappingType(self, map_type):
        assert map_type in ["A", "B"], "The value must be in A or B"
        self._MappingType = map_type

    @property
    def SymbolAllocation(self):
        return self._SymbolAllocation

    @SymbolAllocation.setter
    def SymbolAllocation(self, sym_list):
        # assert prbset > 0 and prbset < self._n_size_bwp, "The value must be 0..n_size_bwp"
        self._SymbolAllocation = sym_list

    @property
    def PRBSet(self):
        return self._PRBSet

    @PRBSet.setter
    def PRBSet(self, prbset):
        assert min(prbset) >= 0 and max(prbset) <= 274, "The value must be 0..274"
        self._PRBSet = prbset

    @property
    def RNTI(self):
        return self._RNTI

    @RNTI.setter
    def RNTI(self, rnti):
        assert rnti >= 0 and rnti <= 65535, "The value must be 0..65535"
        self._RNTI = rnti

    @property
    def EnablePTRS(self):
        return self._EnablePTRS

    @EnablePTRS.setter
    def EnablePTRS(self, enable):
        assert enable in [0, 1], "The value must be 0 or 1"
        self._EnablePTRS = enable


import numpy as np

def subblock_interleaving(u):
    N = u.shape[-1]
    assert np.mod(N,32)==0, \
        "length for sub-block interleaving must be a multiple of 32."
    y = np.zeros_like(u)    
    # Permutation according to Tab 5.4.1.1-1 in 38.212
    perm = np.array([0, 1, 2, 4, 3, 5, 6, 7, 8, 16, 9, 17, 10, 18, 11, 19,
                     12, 20, 13, 21, 14, 22, 15, 23, 24, 25, 26, 28, 27,
                     29, 30, 31])

    for n in range(N):
        i = int(32*n/N)
        j = int(perm[i] * N/32 + np.mod(n, N/32))
        y[n] = u[j]

    return y                        

# nrRateRecoverPolar() in Matlab discards all repeated llr values,
# this is not optimal but we make this function behave the same by default
# optimal behaviour can be switched on by setting discardRepetition to False
def nrRateRecoverPolar(llr, K, N, ibil=False, discardRepetition=True):
    E = llr.shape[0]
    if N <= E:
        # undo repetition
        y = np.copy(llr[:N])
        if discardRepetition == False:
            for k in range(N, E):
                y[k%N] += llr[k]
    else:
        # TODO: implement shortening und puncturing
        print("Error!")
        exit()
    if ibil == True:
        print("Error: interleaving of coded bits is not implemented!")
        exit()
    d = np.zeros(N)
    np.put(d, subblock_interleaving(np.arange(N)), y)
    return d



import numpy as np

def nrDLSCHInfo(tbs, tcr):
    assert tbs >= 0
    assert type(tbs) == int
    assert tcr > 0 and tcr < 1

    bg_info = getBGNInfo(tbs, tcr)
    cb_info = getCBSInfo(bg_info["B"], bg_info["BGN"])

    # Get number of bits (including filler bits) to be encoded by LDPC encoder
    if bg_info["BGN"] == 1:
        N = 66*cb_info["Zc"]
    else:
        N = 50*cb_info["Zc"]

    # Combine information into the output structure
    info = dict()
    info["CRC"]      = bg_info["CRC"]  # CRC polynomial
    info["L"]        = bg_info["L"]    # Number of CRC bits
    info["BGN"]      = bg_info["BGN"]  # Base graph number
    info["C"]        = cb_info["C"]    # Number of code block segments
    info["Lcb"]      = cb_info["Lcb"]  # Number of parity bits per code block
    info["F"]        = cb_info["F"]    # Number of <NULL> filler bits per code block
    info["Zc"]       = cb_info["Zc"]   # Selected lifting size
    info["K"]        = cb_info["K"]    # Number of bits per code block after CBS
    info["N"]        = N              # Number of bits per code block after LDPC coding

    # Modify the output fields if tbs is empty or zero
    if tbs is None:
        info["L"]  = 0
        info["F"]  = 0
        info["Zc"] = 2
        info["K"]  = 0
        info["N"]  = 0

    return info


def getBGNInfo(A, R):
    if A <= 292 or (A <= 3824 and R <= 0.67) or R <= 0.25:
        bgn = 2
    else:
        bgn = 1

    if A > 3824:
        L   = 24
        crc = '24A'
    else:
        L   = 16
        crc = '16'

    info = dict()
    info["BGN"] = bgn
    info["B"]   = A + L
    info["L"]   = L
    info["CRC"] = crc

    return info


def getCBSInfo(B, bgn):
    if bgn == 1:
        Kcb = 8448
    else:
        Kcb = 3840

    # Get number of code blocks and length of CB-CRC coded block
    if B <= Kcb:
        L = 0
        C = 1
        Bd = B
    else:
        L = 24 # Length of the CRC bits attached to each code block
        C = np.ceil(B/(Kcb-L))
        Bd = B+C*L

    # Obtain the number of bits per code block (excluding CB-CRC bits)
    cbz = np.ceil(B/C)

    # Get number of bits in each code block (excluding filler bits)
    Kd = np.ceil(Bd/C)

    # Find the minimum value of Z in all sets of lifting sizes in 38.212
    # Table 5.3.2-1, denoted as Zc, such that Kb*Zc>=Kd
    if bgn == 1:
        Kb = 22
    else:
        if B > 640:
            Kb = 10
        elif B > 560:
            Kb = 9
        elif B > 192:
            Kb = 8
        else:
            Kb = 6

    # Zlist = [2:16 18:2:32 36:4:64 72:8:128 144:16:256 288:32:384];
    Zlist = []
    Zlist += list(range(2, 17))
    Zlist += list(range(18, 33, 2))
    Zlist += list(range(36, 65, 4))
    Zlist += list(range(72, 129, 8))
    Zlist += list(range(144, 257, 16))
    Zlist += list(range(288, 385, 32))

    # Zc = min(Zlist(Kb*Zlist >= Kd));
    # Zc = []
    # for x in Zlist:
    #     if Kb*x >= Kd:
    #         Zc.append(x)
    # Zc = min(Zc)
    Zc = min([x for x in Zlist if (Kb*x >= Kd)])

    # Get number of bits (including <NULL> filler bits) to be input to the LDPC encoder
    if bgn == 1:
        K = 22*Zc
    else:
        K = 10*Zc

    info = dict()
    info["C"]   = C     # Number of code block segments
    info["CBZ"] = cbz   # Number of bits in each code block (excluding CB-CRC bits and filler bits)
    info["Lcb"] = L     # Number of parity bits in each code block
    info["F"]   = K-Kd  # Number of filler bits in each code block
    info["K"]   = K     # Number of bits in each code block (including CB-CRC bits and filler bits)
    info["Zc"]  = Zc    # Selected lifting size
    info["Z"]   = Zlist # Full lifting size set

    return info



from .nrPDSCHDMRSConfig import nrPDSCHDMRSConfig
from .nrPDSCHPTRSConfig import nrPDSCHPTRSConfig

class PDSCHConfigBase():
    def __init__(self):
        self._Modulation = 'qpsk'
        self._NumLayers = 1
        self._ReservedPRB = None
        self._VRBToPRBInterleaving = 0
        self._VRBBundleSize = 2

        self.DMRS = nrPDSCHDMRSConfig()
        self.PTRS = nrPDSCHPTRSConfig()

    @property
    def Modulation(self):
        return self._Modulation

    @Modulation.setter
    def Modulation(self, mod):
        assert mod in ['qpsk', 'qam16', 'qam64', 'qam256'], "The value must be in ['qpsk', 'qam16', 'qam64', 'qam256']"
        self._Modulation = mod

    @property
    def NumLayers(self):
        return self._NumLayers

    @NumLayers.setter
    def NumLayers(self, N):
        assert N in list(range(1, 9)), "The value must be 1..8"
        self._NumLayers = N

    @property
    def ReservedPRB(self):
        raise NotImplementedError('not ready yet')
        return self._ReservedPRB

    @ReservedPRB.setter
    def ReservedPRB(self, N):
        raise NotImplementedError('not ready yet')
        self._ReservedPRB = None

    @property
    def VRBToPRBInterleaving(self):
        raise NotImplementedError('not ready yet')
        return self._VRBToPRBInterleaving

    @VRBToPRBInterleaving.setter
    def VRBToPRBInterleaving(self, N):
        raise NotImplementedError('not ready yet')
        assert N in [0, 1], "The value must be 0 or 1"
        self._VRBToPRBInterleaving = 0

    @property
    def VRBBundleSize(self):
        raise NotImplementedError('not ready yet')
        return self._VRBBundleSize

    @VRBBundleSize.setter
    def VRBBundleSize(self, N):
        raise NotImplementedError('not ready yet')
        assert N in [2, 4], "The value must be 2 or 4"
        self._VRBBundleSize = 0




# 38.211

import numpy as np

from py3gpp.nrPRBS import nrPRBS
from py3gpp.nrSymbolModulate import nrSymbolModulate
from py3gpp.configs.nrPDSCHConfig import nrPDSCHConfig
from py3gpp.configs.nrCarrierConfig import nrCarrierConfig
from py3gpp.nrPDSCHDMRS import PDSCHDMRSSyms

def nrPDSCHPTRS(carrier: nrCarrierConfig, cfg: nrPDSCHConfig):
    if cfg.EnablePTRS == 0:
        return []

    bwp_size = cfg.NRBSize * cfg.NSizeBWP

    # Generates first DMRS symbol (l0)
    n_scid = 0
    dmrs_occupied_syms = PDSCHDMRSSyms(cfg)
    cinit_dmrs = PDSCHPTRScinit(carrier.SymbolsPerSlot, carrier.NSlot, dmrs_occupied_syms[0], cfg.DMRS.NIDNSCID, n_scid)
    dmrs_prbs = nrPRBS(cinit_dmrs, bwp_size)
    dmrs_sym = nrSymbolModulate(dmrs_prbs, "QPSK")

    # Form PTRS symbols from DMRS l0 sequency
    ptrs_prb_set = PTRSPRBSet(cfg)

    ptrs_sym = np.array([dmrs_sym[x] for x in ptrs_prb_set])

    ptrs_occupied_syms = PDSCHPTRSSyms(carrier, cfg)

    return np.tile(ptrs_sym, len(ptrs_occupied_syms))


def PTRSPRBSet(cfg: nrPDSCHConfig):
    # Align with frequency offset based on DMRS type
    if cfg.DMRS.DMRSConfigurationType == 1:
        dmrs_ref_table = np.array([0, 2, 6, 8])
    else:
        dmrs_ref_table = np.array([0, 2, 4, 6])
    kRE_idx = int(cfg.PTRS.REOffset, 2)
    dmrsRef = dmrs_ref_table[kRE_idx]

    Nrb_mod_Kptrs = len(cfg.PRBSet) % cfg.PTRS.FrequencyDensity
    if Nrb_mod_Kptrs == 0:
        kRBref = cfg.RNTI % cfg.PTRS.FrequencyDensity
    else:
        kRBref = cfg.RNTI % Nrb_mod_Kptrs

    if cfg.DMRS.DMRSConfigurationType == 1:
        n_dmrs_per_re = 6
    else:
        n_dmrs_per_re = 4

    prb_set = np.array(cfg.PRBSet[kRBref::cfg.PTRS.FrequencyDensity])
    prb_set = (prb_set*n_dmrs_per_re)+dmrsRef//2

    return prb_set


# LUT for PTRS occupied symbols positions
def PDSCHPTRSSyms(carrier: nrCarrierConfig, cfg: nrPDSCHConfig):
    # First get DMRS symbols
    dmrs_syms = PDSCHDMRSSyms(cfg)
    # Then calculate positions of PTRS
    A_pos = cfg.DMRS.DMRSTypeAPosition
    td = cfg.PTRS.TimeDensity

    ptrs_cnt = td-1
    occupied_syms = np.array([], dtype=int)
    for i in range(cfg.SymbolAllocation[0], carrier.SymbolsPerSlot):
        if i == A_pos:
            ptrs_cnt = 0
            continue

        if i in dmrs_syms:
            ptrs_cnt = 0
            continue

        ptrs_cnt += 1

        if ptrs_cnt == td:
            occupied_syms = np.append(occupied_syms, i)
            ptrs_cnt = 0

    return occupied_syms

def PDSCHPTRScinit(sps, n_slot, n_symb, NIDSCID, n_scid):
    return 2**17 * (sps * n_slot + n_symb + 1) * (2*NIDSCID + 1) + 2*NIDSCID + n_scid




class PDSCHPTRSConfigBase():
    def __init__(self):
        self._TimeDensity = 1
        self._FrequencyDensity = 2
        self._REOffset = '00'

    @property
    def TimeDensity(self):
        return self._TimeDensity

    @TimeDensity.setter
    def TimeDensity(self, td):
        assert td in [1, 2, 4], "The value must be 1, 2 or 4"
        self._TimeDensity = td

    @property
    def FrequencyDensity(self):
        return self._FrequencyDensity

    @FrequencyDensity.setter
    def FrequencyDensity(self, fd):
        assert fd in [2, 4], "The value must be 2 or 4"
        self._FrequencyDensity = fd

    @property
    def REOffset(self):
        return self._REOffset

    @REOffset.setter
    def REOffset(self, offset):
        assert offset in ['00', '01', '10', '11'], "The value must be in ['00', '01', '10', '11']"
        self._REOffset = offset



import numpy as np
from py3gpp.helper import _calc_m_seq

def nrPSS(ncellid):
    assert ncellid >= 0 and ncellid <= 1007, "invalid ncellid"
    N = 7
    c = [0, 1, 1, 0, 1, 1, 1]
    taps = [0, 4]
    m = _calc_m_seq(N, c, taps)
    d_PSS = 1 - 2 * np.roll(m, -nrPSSv(ncellid))
    return d_PSS

def nrPSSv(ncellid):
    n2 = ncellid % 3
    return 43*n2



import os
import sys
import itertools
import numpy as np
import pytest

from py3gpp.nrPDSCHDMRS import nrPDSCHDMRS
from py3gpp.configs.nrPDSCHConfig import nrPDSCHConfig
from py3gpp.configs.nrCarrierConfig import nrCarrierConfig

sys.path.append("test_data")

from test_data.pdsch import pdschdmrs_symbols_ref

def run_nr_pdschdmrs(cfg):
    carrier = nrCarrierConfig();
    carrier.SubcarrierSpacing = 120;
    carrier.CyclicPrefix = 'normal';
    carrier.NSizeGrid = 132;
    carrier.NStartGrid = 0;

    pdsch_cfg = nrPDSCHConfig();
    pdsch_cfg.NSizeBWP = cfg['n_size_bwp']
    pdsch_cfg.NStartBWP = cfg['n_start_bwp']
    pdsch_cfg.MappingType = cfg['MappingType']
    pdsch_cfg.DMRS.DMRSTypeAPosition = cfg['DMRSTypeAPosition']
    pdsch_cfg.DMRS.DMRSLength = cfg['DMRSLength']
    pdsch_cfg.DMRS.DMRSAdditionalPosition = cfg['DMRSAdditionalPosition']
    pdsch_cfg.DMRS.DMRSConfigurationType = cfg['DMRSConfigurationType']
    pdsch_cfg.DMRS.NIDNSCID = cfg['NIDNSCID']
    pdsch_cfg.DMRS.NSCID = cfg['NSCID']
    pdsch_cfg.PRBSet = cfg['PRBSet']
    pdsch_cfg.SymbolAllocation = cfg['SymbolAllocation']

    pdschdmrs_syms = nrPDSCHDMRS(pdsch_cfg, carrier)

    # Cut neccesary part of reference indices
    # TODO: now it works only for 0 and 3 symbols. 1 and 2 quite tricky to cut...
    ref_data = pdschdmrs_symbols_ref
    first_idx = min(pdsch_cfg.PRBSet)*pdsch_cfg.NRBSize
    last_idx = (pdsch_cfg.DMRS.DMRSAdditionalPosition+1) * (max(pdsch_cfg.PRBSet)+1) * pdsch_cfg.NRBSize
    last_idx = last_idx//2
    ref_data = ref_data[first_idx:last_idx]

    pdschdmrs_syms = np.around(pdschdmrs_syms, 4)
    pdschdmrs_syms_ref = np.around(ref_data, 4)

    assert np.array_equal(pdschdmrs_syms, pdschdmrs_syms_ref)


@pytest.mark.parametrize('dmrs_add_pos', [0, 3])
def test_nr_pdschdmrs(dmrs_add_pos):
    cfg = {}
    cfg['n_size_bwp'] = 132
    cfg['n_start_bwp'] = 0
    cfg['MappingType'] = "A"
    cfg['DMRSTypeAPosition'] = 2
    cfg['DMRSLength'] = 1
    cfg['DMRSAdditionalPosition'] = dmrs_add_pos
    cfg['PRBSet'] = list(range(0, 132))
    cfg['SymbolAllocation'] = [2, 12]
    cfg['DMRSConfigurationType'] = 1
    cfg['NIDNSCID'] = 1
    cfg['NSCID'] = 0

    run_nr_pdschdmrs(cfg)



from py3gpp.nrPRBS import nrPRBS
from py3gpp.nrSymbolModulate import nrSymbolModulate


def nrPBCHDMRS(ncellid, ibar_SSB):
    assert ncellid >= 0 and ncellid <= 1007
    assert ibar_SSB >= 0 and ibar_SSB <= 7

    c_init = nrPBCHDMRScinit(ibar_SSB, ncellid)
    c = nrPRBS(c_init, 2*144)

    return nrSymbolModulate(c, 'qpsk')

def nrPBCHDMRScinit(issb, ncellid):
    return 2**11 * (issb + 1) * (ncellid//4 + 1) + 2**6 * (issb + 1) + (ncellid % 4)




# 38.211

import numpy as np

from py3gpp.nrPDSCHPTRS import PDSCHPTRSSyms
from py3gpp.configs.nrPDSCHConfig import nrPDSCHConfig
from py3gpp.configs.nrCarrierConfig import nrCarrierConfig

def nrPDSCHPTRSIndices(carrier: nrCarrierConfig, cfg: nrPDSCHConfig):
    if cfg.EnablePTRS == 0:
        return np.array([])

    frame_begin = cfg.NRBSize * min(cfg.PRBSet)
    frame_end = cfg.NRBSize * (max(cfg.PRBSet)+1)
    frame_size = cfg.NRBSize * cfg.NSizeBWP

    # Align with frequency offset based on DMRS type
    if cfg.DMRS.DMRSConfigurationType == 1:
        kREref_table = [0, 2, 6, 8]
    else:
        kREref_table = [0, 1, 6, 7]
    kREref = kREref_table[int(cfg.PTRS.REOffset, 2)]

    Nrb_mod_Kptrs = len(cfg.PRBSet) % cfg.PTRS.FrequencyDensity
    if Nrb_mod_Kptrs == 0:
        kRBref = cfg.RNTI % cfg.PTRS.FrequencyDensity
    else:
        kRBref = cfg.RNTI % Nrb_mod_Kptrs
    kRBref *= cfg.NRBSize

    # Move to BWP offset and the rest frequency offsets
    ptrs_begin = frame_begin + kREref + kRBref

    # Calculate PTRS positions in every occupied symbol. Frequency position,
    # for every 2 or 4 RBs
    occupied_res = np.array(list(range(ptrs_begin, frame_end, (cfg.PTRS.FrequencyDensity*cfg.NRBSize))))

    # Calculates occupied symbols numbers. Time positions
    occupied_syms = PDSCHPTRSSyms(carrier, cfg)

    # Move PTRS indices to occupied symbols
    ptrs_indices = np.array([occupied_res+(sym*frame_size) for sym in occupied_syms])
    if len(ptrs_indices) > 0:
        ptrs_indices = np.concatenate(ptrs_indices, axis=None)

    return ptrs_indices




# 38.211 5.1 Modulation mapper

import numpy as np

def nrSymbolModulate(databits, modulation):
    int_modtype = modulation.lower()

    assert int_modtype in ["pi/2-bpsk", "bpsk", "qpsk", "16qam", "64qam", "256qam"], "modulation type is incorrect"
    assert len(databits) > 0, "length of databits must be greater 0"

    if int_modtype == "bpsk":
        res_arr = np.zeros(len(databits), dtype=(complex))
        for idx, sample in enumerate(databits):
            res_arr[idx] = (1-2*sample) + 1j*(1-2*sample)
        res_arr = [x/np.sqrt(2) for x in res_arr]

    elif int_modtype == "pi/2-bpsk":
        res_arr = np.zeros(len(databits), dtype=(complex))
        for idx, sample in enumerate(databits):
            if idx%2:
                res_arr[idx] = (2*sample-1) + 1j*(1-2*sample)
            else:
                res_arr[idx] = (1-2*sample) + 1j*(1-2*sample)
        res_arr = [x/np.sqrt(2) for x in res_arr]

    elif int_modtype == "qpsk":
        assert not (len(databits) % 2), "length of databits must be multiple of 2"
        res_arr = np.zeros((len(databits)//2), dtype=(complex))
        chunks = np.array_split(databits, len(databits)//2)
        for idx, sample in enumerate(chunks):
            res_arr[idx] = (1-2*sample[0]) + 1j*(1-2*sample[1])
        res_arr = [x/np.sqrt(2) for x in res_arr]

    elif int_modtype == "16qam":
        assert not (len(databits) % 4), "length of databits must be multiple of 4"
        res_arr = np.zeros((len(databits)//4), dtype=(complex))
        chunks = np.array_split(databits, len(databits)//4)
        for idx, sample in enumerate(chunks):
            res_arr[idx] = ( (1-2*sample[0]) * (2-(1-2*sample[2])) ) + 1j*( (1-2*sample[1]) * (2-(1-2*sample[3])) )
        res_arr = [x/np.sqrt(10) for x in res_arr]

    elif int_modtype == "64qam":
        assert not (len(databits) % 6), "length of databits must be multiple of 6"
        res_arr = np.zeros((len(databits)//6), dtype=(complex))
        chunks = np.array_split(databits, len(databits)//6)
        for idx, sample in enumerate(chunks):
            res_arr[idx] = ( (1-2*sample[0]) * (4-(1-2*sample[2]) * (2-(1-2*sample[4]))) ) + 1j*( (1-2*sample[1]) * (4-(1-2*sample[3]) * (2-(1-2*sample[5]))) )
        res_arr = [x/np.sqrt(42) for x in res_arr]

    elif int_modtype == "256qam":
        assert not (len(databits) % 8), "length of databits must be multiple of 8"
        res_arr = np.zeros((len(databits)//8), dtype=(complex))
        chunks = np.array_split(databits, len(databits)//8)
        for idx, sample in enumerate(chunks):
            res_arr[idx] = ( (1-2*sample[0]) * (8-(1-2*sample[2]) * (4-(1-2*sample[4]) * (2-(1-2*sample[6])))) ) + 1j*( (1-2*sample[1]) * (8-(1-2*sample[3]) * (4-(1-2*sample[5]) * (2-(1-2*sample[7])))) )
        res_arr = [x/np.sqrt(170) for x in res_arr]

    return np.array(res_arr)




import numpy as np
from py3gpp.nrSymbolModulate import nrSymbolModulate
from py3gpp.nrPRBS import nrPRBS

def nrPDSCH(cws, modulation, nlayers, nid, nrnti):
    if nlayers != 1:
        raise NotImplementedError('not ready yet')
    assert nlayers in list(range(1, 9))

    if nlayers > 4:
        ncw = 2
    else:
        ncw = 1

    scrambled_bits = []
    res_sym = []
    for i in range(ncw):
        databits = cws[i]
        cinit = (nrnti * 2**15) + (i * 2**14) + nid
        prbs_bits = nrPRBS(cinit, len(databits))
        scrambled_bits.append(np.bitwise_xor(databits, prbs_bits))
        scrambled_bits = np.array(scrambled_bits)

        res_sym.append(nrSymbolModulate(scrambled_bits[i], modulation[i]))

    # TODO
    # res_sym = nrLayerMap(res_sym, nlayers)
    res_sym = np.hstack(res_sym)

    return np.array(res_sym)



import numpy as np
from py3gpp.helper import _calc_m_seq


def nrSSSm0(ncellid):
    n1 = ncellid // 3
    n2 = ncellid % 3
    return 15*(n1 // 112) + 5*n2


def nrSSSm1(ncellid):
    n1 = ncellid // 3
    return n1 % 112


def _gold(ncellid):
    N = 7
    c = [1, 0, 0, 0, 0, 0, 0]
    taps_0 = [0, 4]
    taps_1 = [0, 1]
    mseq_0 = _calc_m_seq(N, c, taps_0)
    mseq_1 = _calc_m_seq(N, c, taps_1)
    m_0 = nrSSSm0(ncellid)
    m_1 = nrSSSm1(ncellid)
    d_SSS = (1 - 2 * np.roll(mseq_0, -m_0)) * (1 - 2 * np.roll(mseq_1, -m_1))
    return d_SSS


def nrSSS(ncellid):
    assert ncellid >= 0 and ncellid <= 1007, "invalid ncellid"
    return _gold(ncellid)




import numpy as np
from py3gpp.nrOFDMInfo import nrOFDMInfo
from py3gpp.configs.nrCarrierConfig import nrCarrierConfig

# TODO: implement CyclicPrefixFraction
def nrOFDMDemodulate(
    carrier=None,
    waveform=None,
    nrb=None,
    scs=None,
    initialNSlot=None,
    CyclicPrefix="normal",
    Nfft=None,
    SampleRate=None,
    CarrierFrequency=0,
    CyclicPrefixFraction=0.5,
):
    if waveform is None:
        print("Error: no waveform given!")
        return
    if carrier is None:
        if nrb == None:
            print("Error: nrb is needed without carrierConfig!")
            return
        if scs == None:
            print("Error: scs is needed without carrierConfig!")
            return
        if initialNSlot == None:
            print("Error: initialNSlot is needed without carrierConfig!")
            return
        carrier = nrCarrierConfig(1, NSizeGrid = nrb, NStartGrid = 0, SubcarrierSpacing = scs, NSlot = initialNSlot)
    else:
        nrb = carrier.NSizeGrid
        scs = carrier.SubcarrierSpacing
        initialNSlot = 0 if initialNSlot is None else initialNSlot
        
    if Nfft == None:
        if SampleRate == None:
            Nfft = nrOFDMInfo(nrb=nrb, scs=scs)["Nfft"]
            SampleRate = int(Nfft * scs * 1000)
        else:
            Nfft = int(SampleRate // scs // 1000)
    mu = (scs // 15) - 1
    if CyclicPrefix == "normal":
        N_cp1 = int(((144) * 2 ** (-mu) + 16) * (SampleRate / 30720000))
        N_cp2 = int((144 * 2 ** (-mu)) * (SampleRate / 30720000))
    else:
        N_cp1 = int((512 * 2 ** (-mu)) * (SampleRate / 30720000))
        N_cp2 = N_cp1
    N_cp = np.zeros(carrier.SymbolsPerSlot, dtype=int)
    for i in range(len(N_cp)):
        N_cp[i] = N_cp1 if i == 0 or i == 7 * 2 ** (mu) else N_cp2

    idx = 0
    sym_pos_in_slot = initialNSlot
    grid = np.zeros((nrb * 12, 0), "complex")

    sample_pos_in_slot = 0
    for i in range(initialNSlot):
        sample_pos_in_slot += Nfft + N_cp[i]

    symbols_per_slot = carrier.SymbolsPerSlot
    while idx + Nfft <= waveform.shape[0]:
        sym_pos_in_slot = sym_pos_in_slot % symbols_per_slot
        cp_advance = int(CyclicPrefixFraction * N_cp[sym_pos_in_slot])
        idx += cp_advance
        symbol_t = waveform[idx:][:Nfft]
        symbol_f = np.fft.fftshift(np.fft.fft(symbol_t))

        symbol_f *= np.exp(1j*2*np.pi*(N_cp[sym_pos_in_slot] - cp_advance)/Nfft*np.arange(len(symbol_f)))
        symbol_f *= np.exp(1j*np.pi*(N_cp[sym_pos_in_slot] - cp_advance))

        symbol_f = symbol_f[Nfft // 2 - nrb * 12 // 2 : Nfft // 2 + nrb * 12 // 2]

        # phase compensation according to TS 38.211 section 5.4
        if sym_pos_in_slot == 0:
            sample_pos_in_slot = 0
        sample_pos_in_slot += N_cp[sym_pos_in_slot]
        # print(f'symbol {sym_pos_in_slot}, pos {sample_pos_in_slot}, CP {N_cp[sym_pos_in_slot]}, pos {idx - cp_advance}')
        symbol_f *= np.exp(1j * 2 * np.pi * CarrierFrequency / SampleRate * sample_pos_in_slot)
        sample_pos_in_slot += Nfft

        grid = np.concatenate((grid, np.expand_dims(symbol_f, 1)), axis=1)
        idx += Nfft + (N_cp[sym_pos_in_slot] - cp_advance)
        # print(f'slot {slot}, cp_len {N_cp}')
        sym_pos_in_slot = sym_pos_in_slot + 1
    return grid



import numpy as np
from py3gpp.configs.nrCarrierConfig import nrCarrierConfig
from py3gpp.nrOFDMInfo import nrOFDMInfo

# TODO: implement windowing
def nrOFDMModulate(
    carrier=None, grid=None, scs=None, initialNSlot=None, CyclicPrefix="normal", Nfft=None,
    SampleRate=None, Windowing=None, CarrierFrequency=0
):
    info = {}

    if carrier is None:
        if grid is None:
            print("Error: grid is needed when no carrierConfig is specified!")
            return
        NSizeGrid = grid.shape[0] // 12
        carrier = nrCarrierConfig(NSizeGrid=NSizeGrid)
        if initialNSlot is None:
            initialNSlot = carrier.NSlot
    else:
        NSizeGrid = carrier.NSizeGrid
        if initialNSlot is None:
            initialNSlot = carrier.NSlot
        
    if initialNSlot is None:
        initialNSlot = 0

    if scs is None:
        scs = carrier.SubcarrierSpacing

    if Nfft is None:
        if SampleRate is None:
            Nfft = nrOFDMInfo(nrb=NSizeGrid, scs=scs)["Nfft"]
            SampleRate = int(Nfft * scs * 1000)
        else:
            Nfft = int(SampleRate // scs // 1000)

    mu = (scs // 15) - 1

    info["Nfft"] = Nfft
    info["SampleRate"] = SampleRate
    info["CyclicPrefixLengths"] = np.empty(0, "int")
    info["SymbolLengths"] = np.empty(0, "int")
    if len(grid.shape) == 1:
        grid = grid[..., np.newaxis]
    nSlots = grid.shape[1]
    waveform = np.empty(0, "complex")
    if CyclicPrefix == "normal":
        N_cp1 = int(((144) * 2 ** (-mu) + 16) * (SampleRate / 30720000))
        N_cp2 = int((144 * 2 ** (-mu)) * (SampleRate / 30720000))
    else:
        N_cp1 = int((512 * 2 ** (-mu)) * (SampleRate / 30720000))
        N_cp2 = N_cp1
    N_cp = np.zeros(carrier.SymbolsPerSlot, dtype=int)
    for i in range(len(N_cp)):
        N_cp[i] = N_cp1 if i == 0 or i == 7 * 2 ** (mu) else N_cp2

    sample_pos_in_slot = 0
    for i in range(initialNSlot):
        sample_pos_in_slot += Nfft + N_cp[i]
        
    for sym_pos_in_grid in range(nSlots):
        sym_pos_in_slot = (sym_pos_in_grid + initialNSlot) % carrier.SymbolsPerSlot
        symbol_len = Nfft + N_cp[sym_pos_in_slot]
        nFill = (Nfft - grid.shape[0]) // 2
        if nFill < 0:
            full_slot_grid = grid[np.abs(nFill) :, sym_pos_in_grid][:nFill]
        else:
            full_slot_grid = np.concatenate([np.zeros(nFill), grid[:, sym_pos_in_grid], np.zeros(nFill)])

        # phase compensation according to TS 38.211 section 5.4
        if sym_pos_in_slot == 0:
            sample_pos_in_slot = 0
        sample_pos_in_slot += N_cp[sym_pos_in_slot]
        # print(f'symbol {sym_pos_in_slot}, pos {sample_pos_in_slot}, CP {N_cp[sym_pos_in_slot]}, pos = {len(waveform)}')
        full_slot_grid *= np.exp(-1j * 2 * np.pi * CarrierFrequency / SampleRate * sample_pos_in_slot)
        sample_pos_in_slot += Nfft
        
        symbol_waveform = np.fft.ifft(np.fft.fftshift(full_slot_grid))
        symbol_waveform_cp = np.append(symbol_waveform[-N_cp[sym_pos_in_slot]:], symbol_waveform)
        waveform = np.append(waveform, symbol_waveform_cp)
        info["CyclicPrefixLengths"] = np.append(info["CyclicPrefixLengths"], N_cp[sym_pos_in_slot])
        info["SymbolLengths"] = np.append(info["SymbolLengths"], symbol_len)
        # print(f'slot {slot_num}, cp_len {N_cp[sym_pos_in_slot]}')
    return [waveform, info]



import os
import itertools
import numpy as np
import pytest
import sys

from py3gpp.nrPDSCHPTRS import PDSCHPTRSSyms
from py3gpp.nrPDSCHPTRSIndices import nrPDSCHPTRSIndices
from py3gpp.configs.nrPDSCHConfig import nrPDSCHConfig
from py3gpp.configs.nrCarrierConfig import nrCarrierConfig

sys.path.append("test_data")
from test_data.pdsch import pdschptrs_indices_ref

def run_nr_pdschptrs(cfg):
    carrier = nrCarrierConfig();
    carrier.SubcarrierSpacing = 120;
    carrier.CyclicPrefix = 'normal';
    carrier.NSizeGrid = 132;
    carrier.NStartGrid = 0;

    pdsch_cfg = nrPDSCHConfig();
    pdsch_cfg.NSizeBWP = cfg['n_size_bwp']
    pdsch_cfg.NStartBWP = cfg['n_start_bwp']
    pdsch_cfg.MappingType = cfg['MappingType']
    pdsch_cfg.DMRS.DMRSTypeAPosition = cfg['DMRSTypeAPosition']
    pdsch_cfg.DMRS.DMRSLength = cfg['DMRSLength']
    pdsch_cfg.DMRS.DMRSAdditionalPosition = cfg['DMRSAdditionalPosition']
    pdsch_cfg.DMRS.DMRSConfigurationType = cfg['DMRSConfigurationType']
    pdsch_cfg.DMRS.NIDNSCID = cfg['NIDNSCID']
    pdsch_cfg.DMRS.NSCID = cfg['NSCID']
    pdsch_cfg.RNTI = cfg['RNTI']
    pdsch_cfg.PRBSet = cfg['PRBSet']
    pdsch_cfg.SymbolAllocation = cfg['SymbolAllocation']

    pdsch_cfg.EnablePTRS = cfg['EnablePTRS']
    pdsch_cfg.PTRS.TimeDensity = cfg['PTRSTimeDensity']
    pdsch_cfg.PTRS.FrequencyDensity = cfg['PTRSFrequencyDensity']
    pdsch_cfg.PTRS.REOffset = cfg['PTRSREOffset']

    pdschptrs_indices = nrPDSCHPTRSIndices(carrier, pdsch_cfg)

    # Align MATLAB with Python
    ref_indices = pdschptrs_indices_ref - 1

    assert np.array_equal(pdschptrs_indices, ref_indices)


def test_nr_pdschptrs():
    cfg = {}
    cfg['n_size_bwp'] = 132
    cfg['n_start_bwp'] = 0
    cfg['MappingType'] = "A"
    cfg['DMRSTypeAPosition'] = 3
    cfg['DMRSLength'] = 1
    cfg['DMRSAdditionalPosition'] = 3
    cfg['PRBSet'] = list(range(0, 132))
    cfg['SymbolAllocation'] = [2, 12]
    cfg['DMRSConfigurationType'] = 2
    cfg['NIDNSCID'] = 1
    cfg['NSCID'] = 0
    cfg['RNTI'] = 1
    cfg['EnablePTRS'] = 1
    cfg['PTRSTimeDensity'] = 2
    cfg['PTRSFrequencyDensity'] = 4
    cfg['PTRSREOffset'] = '11'

    run_nr_pdschptrs(cfg)

if __name__ == '__main__':
    test_nr_pdschptrs(1)



class nrNumerologyConfig:
        def __init__(self):
            self._SubcarrierSpacing = 15
            self._CyclicPrefix = "normal"

        @property
        def SubcarrierSpacing(self):
            return self._SubcarrierSpacing

        @SubcarrierSpacing.setter
        def SubcarrierSpacing(self, spacing):
            assert spacing in [15, 30, 60, 120, 240], "The value must be [15, 30, 60, 120, 240]"
            self._SubcarrierSpacing = spacing

        @property
        def CyclicPrefix(self):
            return self._CyclicPrefix

        @CyclicPrefix.setter
        def CyclicPrefix(self, pref):
            assert pref in ["normal", "extended"], "The value must be in ['extended' or 'normal'']"
            self._CyclicPrefix = pref


import numpy as np
import pytest

from py3gpp.nrOFDMModulate import nrOFDMModulate
from py3gpp.nrOFDMDemodulate import nrOFDMDemodulate
from py3gpp.configs.nrCarrierConfig import nrCarrierConfig

@pytest.mark.parametrize("f0_hz", [0, 10e6, 11e6])
@pytest.mark.parametrize("initial_N_slot", [0, 4])
def test_phase_compensation(f0_hz, initial_N_slot):
    carrier = nrCarrierConfig(1, NSizeGrid = 30)
    Nfft = 512
    num_slots = 20
    expected_waveform_len = Nfft * num_slots + 36 * 17 + 40 * 3
    print(f'testing f0 = {f0_hz} Hz, initialNSlot = {initial_N_slot}')
    test_symbol = np.ones((carrier.NSizeGrid * 12, 1), dtype = "complex")
    test_symbols = np.repeat(test_symbol, num_slots)
    test_symbols = np.reshape(test_symbols, (carrier.NSizeGrid * 12, num_slots))
    waveform, _ = nrOFDMModulate(carrier, test_symbols, CarrierFrequency = f0_hz, initialNSlot = initial_N_slot)
    assert waveform.shape[0] == expected_waveform_len
    rx_test_symbols = nrOFDMDemodulate(carrier, waveform, CarrierFrequency = f0_hz, initialNSlot = initial_N_slot)
    assert np.all(test_symbols == np.round(rx_test_symbols))
    
if __name__ == '__main__':
     test_phase_compensation(f0_hz = 10e6, initial_N_slot = 4)


from .nrNumerologyConfig import nrNumerologyConfig

class nrCarrierConfig(nrNumerologyConfig):
        def __init__(self, NCellID = 1, NSizeGrid = 52, NStartGrid = 0, NSlot = 0, NFrame = 0, SubcarrierSpacing = 15):
            super().__init__()
            self._NCellID = NCellID
            self._NSizeGrid = NSizeGrid
            self._NStartGrid = NStartGrid
            self._NSlot = NSlot
            self._NFrame = NFrame
            self.SubcarrierSpacing = SubcarrierSpacing

        @property
        def SymbolsPerSlot(self):
            if self.CyclicPrefix == "normal":
                return 14
            else:
                return 12

        @property
        def SlotsPerSubframe(self):
            return self.SubcarrierSpacing//15

        @property
        def SlotsPerFrame(self):
            return int(10*(self.SubcarrierSpacing/15))

        @property
        def NCellID(self):
            return self._NCellID

        @NCellID.setter
        def NCellID(self, nid):
            assert nid >= 0 and nid <= 1007, "The value must be in the range 0...1007."
            self._NCellID = nid

        @property
        def NSizeGrid(self):
            return self._NSizeGrid

        @NSizeGrid.setter
        def NSizeGrid(self, size):
            assert size > 0 and size <= 275, "The value must be in the range 1...275."
            self._NSizeGrid = size

        @property
        def NStartGrid(self):
            return self._NStartGrid

        @NStartGrid.setter
        def NStartGrid(self, start):
            assert start >= 0 and start <= 2199, "The value must be in the range 0...2999."
            self._NStartGrid = start

        @property
        def NSlot(self):
            return self._NSlot

        @NSlot.setter
        def NSlot(self, num):
            assert num >= 0, "The value must be scalar, nonnegative."
            self._NSlot = num

        @property
        def NFrame(self):
            return self._NFrame

        @NFrame.setter
        def NFrame(self, num):
            assert num >= 0, "The value must be scalar, nonnegative."
            self._NFrame = num